#!/usr/local/bin/env python
from __future__ import division
from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from radd import theta
from IPython.display import HTML

def render_animation(anim):
    plt.close(anim._fig)
    return HTML(anim_to_html(anim))

def animate_dpm(model):
    """ to render animation within a notebook :
        vis.render_animation(vis.animated_dpm_example(MODEL))
    """
    from matplotlib import animation
    params = deepcopy(model.inits)
    bound=theta.scalarize_params(params)['a']
    x, gtraces, straces, xi, yi, nframes = gen_re_traces(model, params)
    f, axes = build_decision_axis(onset=x[0][0], bound=bound)
    glines = [axes[i].plot([], [], linewidth=1.5)[0] for i, n in enumerate(gtraces)]
    slines = [axes[i].plot([xi[i]], [yi[i]], linewidth=1.5)[0] for i, n in enumerate(straces)]
    f_args = (x, gtraces, glines, straces, slines, params, xi, yi)
    anim=animation.FuncAnimation(f, re_animate_multiax, fargs=f_args, frames=nframes, interval=1, blit=True)
    return anim

def gen_re_traces(model, params):
    sim = deepcopy(model.opt.simulator)
    sim.__update_steps__(dt=.001)
    ssd, nssd, nss, nss_per, ssd_ix = sim.ssd_info
    nsstot = nss * nssd
    params = sim.vectorize_params(params)
    bound = params['a'][0]
    Pg, Tg, xtb = sim.__update_go_process__(params)
    Ps, Ts = sim.__update_stop_process__(params)
    Ts=[Ts[0, i] for i in [2, -1]]
    dvg, dvs = sim.sim_fx(params, analyze=False)
    dvg = dvg[0, :nss, :].reshape(nssd, nss_per, dvg.shape[-1])
    gtraces = [dvg[i, 0] for i in [2, -1]]
    straces = [dvs[0, i, 0] for i in [2, -1]]
    gtraces = [gt[gt <= bound] for gt in gtraces]
    ssi, xinit_ss, integrated, dvgs, dvss = [], [], [], [], []
    for i, (g, s) in enumerate(zip(gtraces, straces)):
        xinit_ss.append(Tg[0] - Ts[i])
        ssi.append(g[:xinit_ss[i]])
        ss = s[:Ts[i]]
        s = np.append(g[:xinit_ss[i]], ss[ss >= 0])
        ixmin = np.min([len(g), len(s)])
        dvgs.append(g[:ixmin])
        dvss.append(s[:ixmin])
    nframes = [len(gt) for gt in dvgs]
    x = params['tr'][0] * 1000 + [np.arange(nf) for nf in nframes]
    sim.__update_steps__(dt=.005)
    return [x, dvgs, dvss, xinit_ss, ssi, np.max(nframes)]

def build_decision_axis(onset, bound, ssd=np.array([300, 400]), tb=650):
    # init figure, axes, properties
    f, axes = plt.subplots(len(ssd), 1, figsize=(7, 7), dpi=300)
    #f.subplots_adjust(wspace=.05, top=.1, bottom=.1)
    f.subplots_adjust(hspace=.05, top=.99, bottom=.09)
    w = tb + 40
    h = bound
    start = onset - 80
    for i, ax in enumerate(axes):
        plt.setp(ax, xlim=(start - 1, w + 1), ylim=(0 - (.01 * h), h + (.01 * h)))
        ax.vlines(x=ssd[i], ymin=0, ymax=h, color="#e74c3c", lw=2.5, alpha=.5)
        ax.hlines(y=h, xmin=start, xmax=w, color='k')
        ax.hlines(y=0, xmin=start, xmax=w, color='k')
        ax.vlines(x=tb, ymin=0, ymax=h, color='#2043B0', lw=2.5, linestyle='-', alpha=.5)
        ax.vlines(x=start + 2, ymin=0, ymax=h, color='k')
        ax.text(ssd[i] + 10, h * .88, str(ssd[i]) + 'ms', fontsize=19)
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_xticks([])
        ax.set_yticks([])
    sns.despine(top=True, right=True, bottom=True, left=True)
    return f, axes

def re_animate_multiax(i, x, gtraces, glines, straces, slines, params, xi, yi):
    gcolor = '#27ae60'
    scolor = '#e74c3c'
    for n in range(len(x)):
        ex, gt, gl, st, sl, ix, iy = [xx[n] for xx in [x, gtraces, glines, straces, slines, xi, yi]]
        try:
            gl.set_data(ex[:i + 1], gt[:i + 1])
            gl.set_color(gcolor)
            sl.set_data(ex[ix:i + 1], st[ix:i + 1])
            sl.set_color(scolor)
        except Exception:
            return sl, gl
        #f.savefig('animation_frames/movie/img' + str(i) +'.png', dpi=300)
    return sl, gl

def anim_to_html(anim):
    from tempfile import NamedTemporaryFile
    VIDEO_TAG = """<video controls>
             <source src="data:video/x-m4v;base64,{0}" type="video/mp4">
             Your browser does not support the video tag.
            </video>"""
    if not hasattr(anim, '_encoded_video'):
        with NamedTemporaryFile(suffix='.mp4') as f:
            anim.save(f.name, extra_args=['-vcodec', 'libx264', '-pix_fmt', 'yuv420p'])
            video = open(f.name, "rb").read()
        anim._encoded_video = video.encode("base64")
    return VIDEO_TAG.format(anim._encoded_video)



#!/usr/local/bin/env python
from __future__ import division
import sys
from future.utils import listvalues
from copy import deepcopy
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from radd.tools import colors, messages, analyze
from radd import theta
from IPython.display import HTML, Javascript, display

class PBinJ(object):
    """ animated ProgressBar (PB) to be used (inJ)upyter notebooks
    (set bartype to 'uglybar' if running from terminal)
    """
    def __init__(self, bartype='colorbar', n=None, color='blue', title='Progress'):
        colors = {'green': '#16a085', 'blue': '#4168B7', 'red': "#e74c3c"}
        self.color = colors[color]
        self.n=n
        self.displayed=False
        self.bartype=bartype
        self.title=title
        self.init_bars()

    def init_bars(self):
        if self.bartype=='uglybar':
            import progressbar
            if self.n is not None:
                self.bar = progressbar.ProgressBar(0, self.n)
            self.new_prog_string = ''.join([self.title, ': {0}'])
            self.update = self.update_uglybar
        else:
            import uuid
            self.barid=str(uuid.uuid4())
            if self.bartype=='colorbar':
                args = [self.title, self.color, "500", self.barid, '0%', 'left']
                self.update = self.update_colorbar
            else:
                args = [self.title, self.color, "108", self.barid, '100%', 'center']
                self.update = self.update_progress
            self.bar="""<div<p>{0}</p> <div style="border: 2px solid {1}; width:{2}px">
            <div id="{3}" style="background-color:{1}; width:{4}; text:''; color:#fff; text-align:{5};">
            &nbsp;</div> </div> </div>""".format(*args)

    def display_bars(self):
        if self.bartype=='uglybar':
            self.bar.start()
        else:
            display(HTML(self.bar))
        self.displayed=True

    def update_progress(self, new_progress=None):
        if self.displayed==False:
            self.display_bars()
        if type(new_progress) is str:
            display(Javascript("$('div#{}').text({})".format(self.barid, new_progress)))
        else:
            display(Javascript("$('div#{}').text({:.5f})".format(self.barid, new_progress)))

    def update_colorbar(self, i=None, new_progress=None):
        if self.displayed==False:
            self.display_bars()
        if i is not None:
            display(Javascript("$('div#{}').width('{:.2f}%')".format(self.barid, ((i+1)*1./(self.n))*100)))
        if new_progress is not None:
            self.update_progress(new_progress)

    def update_uglybar(self, i=None, new_progress=None):
        if self.displayed==False:
            self.display_bars()
        if new_progress is not None:
            sys.stdout.write('\r'+self.new_prog_string.format(str(new_progress)))
            sys.stdout.flush()
        if i is not None:
            self.bar.update(i)

    def clear(self):
        if self.bartype=='uglybar':
            sys.stdout.flush
        else:
            from IPython.display import clear_output
            clear_output()


class NestedProgress(object):
    """ initialize multiple progress bars for tracking nested stages of fitting routine
    """
    def __init__(self, name='inits_bar', bartype='colorbar', n=None, init_state=None, color='blue', title='global fmin'):
        self.bars = {}
        self.history = []
        self.add_bar(name=name, bartype=bartype, n=n, init_state=init_state, color=color, title=title)

    def add_bar(self, name='inits_bar', bartype='colorbar', n=None, init_state=None, color='blue', title='global fmin'):
        bar = PBinJ(bartype=bartype, n=n, color=color, title=title)
        self.bars[name] = bar
        if init_state is not None:
            self.reset_bar(name, init_state)

    def reset_bar(self, name, init_state=None):
        self.history = [init_state]
        self.bars[name].update(new_progress=init_state)

    def update(self, name='all', i=None, new_progress=None):
        if name=='all':
            update_list = listvalues(self.bars)
        else:
            update_list = [self.bars[name]]
        for bar in update_list:
            if bar.bartype=='infobar' and new_progress:
                bar.update(new_progress)
            elif bar.bartype=='colorbar':
                bar.update(i=i, new_progress=new_progress)

    def callback(self, x, fmin, accept):
        """ A callback function for reporting basinhopping status
        Arguments:
            x (array):
                parameter values
            fmin (float):
                function value of the trial minimum, and
            accept (bool):
                whether or not that minimum was accepted
        """
        if fmin <= np.min(self.history):
            self.bars['glb_basin'].update(new_progress=fmin)
        if accept:
            self.history.append(fmin)
            self.bars['lcl_basin'].update(new_progress=fmin)

    def clear(self):
        for bar in listvalues(self.bars):
            bar.clear()


def rwr(X, get_index=False, n=None):
    """
    Modified from http://nbviewer.ipython.org/gist/aflaxman/6871948
    """
    if isinstance(X, pd.Series):
        X = X.copy()
        X.index = range(len(X.index))
    if n == None:
        n = len(X)
    resample_i = np.floor(np.random.rand(n) * len(X)).astype(int)
    X_resample = (X[resample_i])
    if get_index:
        return resample_i
    else:
        return X_resample

def resample_data(data, n=120, groups=['ssd']):
    """ generates n resampled datasets using rwr()
    for bootstrapping model fits
    """
    df = data.copy()
    bootlist = list()
    if n == None:
        n = len(df)
    for level, level_df in df.groupby(groups):
        boots = level_df.reset_index(drop=True)
        orig_ix = np.asarray(boots.index[:])
        resampled_ix = rwr(orig_ix, get_index=True, n=n)
        bootdf = level_df.irow(resampled_ix)
        bootlist.append(bootdf)
    # concatenate and return all resampled conditions
    return self.model.rangl_data(pd.concat(bootlist))

def extract_popt_fitinfo(finfo=None, plist=None, pc_map=None):
    """ takes optimized dict or DF of vectorized parameters and
    returns dict with only depends_on.keys() containing vectorized vals.
    Is accessed by fit.Optimizer objects after optimization routine.
    ::Arguments::
    finfo (dict/DF):
        finfo is dict if self.fit_on is 'average'
        and DF if self.fit_on is 'subjects' or 'bootstrap'
        contains optimized parameters
    ::Returns::
    popt (dict):
        dict with only depends_on.keys() containing
        vectorized vals
    """
    finfo = dict(deepcopy(finfo))
    plist = list(inits)
    popt = {pkey: finfo[pkey] for pkey in plist}
    for pkey in list(pc_map):
        popt[pkey] = np.array([finfo[pc] for pc in pc_map[pkey]])
    return popt

def params_io(p={}, io='w', iostr='popt'):
    """ read // write parameters dictionaries
    """
    if io == 'w':
        pd.Series(p).to_csv(''.join([iostr, '.csv']))
    elif io == 'r':
        ps = pd.read_csv(''.join([iostr, '.csv']), header=None)
        p = dict(zip(ps[0], ps[1]))
        return p

def fits_io(fitparams, fits=[], io='w', iostr='fits'):
    """ read // write y, wts, yhat arrays
    """
    y = fitparams['y'].flatten()
    wts = fitparams['wts'].flatten()
    fits = fits.flatten()
    if io == 'w':
        index = np.arange(len(fits))
        df = pd.DataFrame({'y': y, 'wts': wts, 'yhat': fits}, index=index)
        df.to_csv(''.join([iostr, '.csv']))
    elif io == 'r':
        df = pd.read_csv(''.join([iostr, '.csv']), index_col=0)
        return df



#!usr/bin/env python
from __future__ import division
import os
from future.utils import listvalues
from copy import deepcopy
import pandas as pd
import numpy as np
from numpy import array
from radd.tools import analyze
from scipy.stats.mstats import mquantiles as mq
from scipy.stats.mstats_extras import mjci

class DataHandler(object):

    def __init__(self, model, max_wt=2.5):
        self.model = model
        self.data = model.data
        self.inits = model.inits
        self.model_id = model.model_id
        self.idx = model.idx
        self.nidx = model.nidx
        self.max_wt = max_wt
        self.ssd_method = model.ssd_method
        self.kind = model.kind
        self.fit_on = model.fit_on
        self.quantiles = model.quantiles
        self.groups = model.groups
        self.depends_on = model.depends_on
        self.conds = model.conds
        self.nconds = model.nconds
        self.nlevels = model.nlevels
        self.nrows = self.nidx * self.nlevels * self.nconds
        self.grpData = self.data.groupby(self.groups)

    def make_dataframes(self):
        """ Generates the following dataframes and arrays:
        observed (list of ndarrays):
            list containing ndarrays (ncond x n_observed_data_points) entered into thec
            costfx during fitting. If self.fit_on=='average', self.observed will
            contain a single ndarray (ncond x average_data). If self.fit_on=='subjects',
            self.observed will contain 1 ndarray for each subject and fits will be performed
            iteratively
        observedDF (DF):
              Contains Prob and RT quant. for each subject
              used to calc. cost fx weights
        fits (DF):
              empty DF shaped like observed DF, used to store simulated
              predictions of the optimized model
        fitinfo (DF):
              stores all opt. parameter values and model fit statistics
        """
        self.make_observed_groupDFs()
        odf = self.observedDF.copy()
        wdf = self.wtsDF.copy()
        condvalues = lambda df: df.loc[:, 'acc':].dropna(axis=1).values.squeeze()
        flatvalues = lambda df: df.loc['acc':].values.squeeze()
        if self.fit_on=='subjects':
            self.observed = [condvalues(odf[odf['idx']==idx]) for idx in self.idx]
            self.cond_wts = [condvalues(wdf[wdf['idx']==idx]) for idx in self.idx]
            self.observed_flat = [flatvalues(odf[odf['idx']==idx].mean()) for idx in self.idx]
            self.flat_wts = [flatvalues(wdf[wdf['idx']==idx].mean()) for idx in self.idx]
        elif self.fit_on=='average':
            self.observed = [condvalues(odf.groupby(self.conds).mean())]
            self.cond_wts = [condvalues(wdf.groupby(self.conds).mean())]
            self.observed_flat = [flatvalues(odf.mean())]
            self.flat_wts = [flatvalues(wdf.mean())]

    def make_observed_groupDFs(self):
        """ concatenate all idx data vectors into a dataframe
        """
        masterDF_header = self.__make_headers__()
        data = self.data
        ncols = len(masterDF_header)
        nrows = self.nrows
        index = range(nrows)
        nan_data = np.zeros((nrows, ncols))*np.nan

        self.datdf = self.grpData.apply(self.rangl_data).sortlevel(0)
        self.dfvals = [self.datdf.values[i].astype(float) for i in index]
        self.observedDF = pd.DataFrame(nan_data, columns=masterDF_header, index=index)
        self.observedDF.loc[:, self.groups] = self.datdf.reset_index()[self.groups].values
        # make yhatDF to fill w/ model-predicted data arrays
        self.yhatDF = self.observedDF.copy()
        # make wtsDF for handling cost-fx weights
        self.wtsDF = self.observedDF.copy()
        # make fitDF for storing w/ goodness-of-fit stats and popt
        self.fitDF = pd.DataFrame(columns=self.f_cols, index=range(self.nidx))
        self.fitDF['idx'] = self.idx

        for rowi in range(nrows):
            # fill observedDF one row at a time, using idx_rows
            self.observedDF.loc[rowi, self.idx_cols[rowi]] = self.dfvals[rowi]
        if self.model.weighted:
            # Calculate p(resp) and rt quantile costfx weights
            idx_qwts, idx_pwts = self.estimate_cost_weights()
            # fill wtsDF with idx quantile wts (ratios)
            self.wtsDF.loc[:, self.q_cols] = idx_qwts
            # fill wtsDF with resp. probability wts (ratios)
            self.wtsDF.loc[:, self.p_cols] = idx_pwts
        else:
            self.wtsDF.loc[:, self.p_cols+self.q_cols] = 1
        if self.fit_on=='average':
            observed_err = self.observedDF.groupby(self.conds).sem()*2
            self.observed_err = observed_err.loc[:, 'acc':].values.squeeze()
        else:
            self.varDF=None

    def rangl_data(self, data):
        """ called by __make_dataframes__ to generate
        observed data arrays
        """
        gac = data.query('ttype=="go"').acc.mean()
        grt = data.query('response==1 & acc==1').rt.values
        ert = data.query('response==1 & acc==0').rt.values
        gq = mq(grt, prob=self.model.quantiles)
        eq = mq(ert, prob=self.model.quantiles)
        data_vector = [gac, gq, eq]
        if 'ssd' in self.data.columns:
            stopdf = data.query('ttype=="stop"')
            if self.model.ssd_method=='all':
                sacc=stopdf.groupby('ssd').mean()['acc'].values
            elif self.model.ssd_method=='central':
                sacc = np.array([stopdf.mean()['acc']])
            data_vector.insert(1, sacc)
        return np.hstack(data_vector)

    def fill_fitDF(self, data, fitparams=None):
        """ fill fitDF with fit statistics
        ::Arguments::
            data (Series):
                fitinfo Series containing model statistics and
                optimized parameters (see Model.assess_fit() method)
            fitparams (dict):
                model.fitparams dict w/ meta info for last fit
        """
        if fitparams is None:
            fitparams = self.model.fitparams
        for k in data.keys():
            if '_' in k:
                data[k.split('_')[-1]] = data[k]
                data.drop(k, inplace=True)
        fitDF = self.fitDF.copy()
        row = np.argmax(fitDF['AIC'].isnull())
        if self.fit_on=='average':
            idxname = 'average'
            if self.model.is_nested:
                idxname = self.model.model_id.split('_')[1]
        else:
            idxname = self.idx[fitparams['idx']]
        data_widx = data.copy()
        data_widx['idx'] = idxname
        fitDF.loc[row, self.f_cols] = data_widx
        self.fitDF = fitDF.copy()
        self.model.fitDF = fitDF.copy()

    def fill_yhatDF(self, data, fitparams=None):
        """ fill yhatDF with model predictions
        ::Arguments::
            data (ndarray):
                array containing model predictions (nlevels x ncols)
                where ncols is number of data columns in observedDF
            fitparams (dict):
                model.fitparams dict w/ meta info for last fit
        """
        if fitparams is None:
            fitparams = self.model.fitparams
        yhatDF = self.yhatDF.copy()
        nl = fitparams['nlevels']
        data = data.reshape(nl, int(data.size/nl))
        next_row = np.argmax(yhatDF.isnull().any(axis=1))
        keys = self.idx_cols[next_row]
        for i in range(nl):
            row = next_row+i
            data_series = pd.Series(data[i], index=keys)
            yhatDF.loc[row, keys] = data_series
            if self.fit_on=='average':
                if self.model.is_nested:
                    idxname = self.model.model_id
                else:
                    idxname = 'average'
                yhatDF.loc[row, 'idx'] = idxname
        self.yhatDF = yhatDF.copy()
        self.model.yhatDF = yhatDF.copy()

    def determine_ssd_method(self, stopdf):
        ssd_n = [df.size for _, df in stopdf.groupby('ssd')]
        # test if equal # of trials per ssd & return ssd_n
        all_equal_counts = ssd_n[1:] == ssd_n[:-1]
        if all_equal_counts:
            self.model.ssd_method = 'all'
        else:
            self.model.ssd_method = 'central'
        return self.model.ssd_method

    def set_model_ssds(self, stopdf, index=['idx'], scale=.001):
        """ set model attr "ssd" as list of np.arrays
        ssds to use when simulating data during optimization
        """
        if self.ssd_method is None:
            self.ssd_method = self.determine_ssd_method(stopdf)
        if self.ssd_method == 'all':
            get_df_ssds = lambda df: df.groupby(self.conds).ssd.unique().values
            cond_ssds =  [get_df_ssds(df) for _,df in stopdf.groupby('idx')]
        elif self.ssd_method == 'central':
            mean_cond_ssd_df = stopdf.pivot_table('ssd', index='idx', columns=self.conds)
            cond_ssds = list(mean_cond_ssd_df.values)
        self.model.ssd = [np.sort(np.vstack(ssds))*scale for ssds in cond_ssds]

    def estimate_cost_weights(self):
        """ calculate weights using observed variability
        across subjects (model.observedDF)
        """
        data = self.data
        nsplits = self.nlevels * self.nconds
        percents = self.model.quantiles
        nquant = percents.size
        # estimate quantile weights
        idx_qwts = self.idx_quant_weights()
        # estimate resp. probability weights
        idx_pwts = self.idx_acc_weights()
        return idx_qwts, idx_pwts

    def idx_quant_weights(self):
        """ calculates weight vectors for reactive RT quantiles by
        first estimating the SEM of RT quantiles for corr. and err. responses.
        (using Maritz-Jarrett estimatation: scipy.stats.mstats_extras.mjci).
        Then representing these variances as ratios.
        e.g.
              QSEM = mjci(rtvectors)
              wts = median(QSEM)/QSEM
        """
        idx_mjci = lambda x: mjci(x.rt, prob=self.model.quantiles)
        nidx = self.nidx
        nquant = self.model.quantiles.size
        groups = self.groups
        nsplits = self.nlevels * self.nconds
        # get all trials with response recorded
        godf = self.data.query('response==1')
        # sort by ttype first so go(acc==1) occurs before stop(acc==0)
        ttype_ordered_groups = np.hstack([groups, 'ttype', 'acc']).tolist()
        godf_grouped = godf.groupby(ttype_ordered_groups)
        # apply self.idx_mjci() to estimate quantile CI's
        quant_err = np.vstack(godf_grouped.apply(idx_mjci).values)
        # reshape [nidx   x   ncond * nquant * nacc]
        idx_qerr = quant_err.reshape(nidx, nsplits * nquant * 2)
        # calculate subject median across all conditions quantiles and accuracy
        # this implicitly accounts for n_obs as the mjci estimate of sem will be lower
        # for conditions with greater number of observations (i.e., more acc==1 in godf)
        idx_medians = np.nanmedian(idx_qerr, axis=1)
        idx_qratio = idx_medians[:, None] / idx_qerr
        # set extreme values to max_wt arg val
        idx_qratio[idx_qratio >= self.max_wt] = self.max_wt
        # reshape to fit in wtsDF[:, q_cols]
        return idx_qratio.reshape(nidx * nsplits, nquant * 2)

    def idx_acc_weights(self, index=['idx']):
        """ count number of observed responses across levels, transform into ratios
        (counts_at_each_level / np.median(counts_at_each_level)) for weight
        subject-level p(response) values in cost function.
        ::Arguments::
            df (DataFrame): group-level dataframe
            var (str): column header for variable to count responses
            conds (list): depends_on.values()
        """
        df = self.data.copy()
        if not self.model.is_flat:
            index = index + self.conds
        if 'ssd' in df.columns:
            if self.ssd_method=='all':
                df = df[df.ttype=='stop'].copy()
                split_by = 'ssd'
            else:
                split_by = 'ttype'
        else:
            split_by = self.conds
            _ = index.remove(split_by)
        df['n'] = 1
        countdf = df.pivot_table('n', index=index, columns=split_by, aggfunc=np.sum)
        idx_pwts = countdf.values / countdf.median(axis=1).values[:, None]
        if self.ssd_method=='all':
            go_wts = np.ones(countdf.shape[0])
            idx_pwts = np.concatenate((go_wts[:,None], idx_pwts), axis=1)
        return idx_pwts

    def __make_headers__(self, ssd_list=None):
        g_cols = self.groups
        if 'ssd' in self.data.columns:
            # get ssd's for fits if in datacols
            if 'ssd' in self.data.columns:
                stopdf = self.data[self.data.ttype=='stop']
                self.set_model_ssds(stopdf)
            if self.ssd_method=='all':
                get_df_ssds = lambda df: df.ssd.unique()
                ssds = [get_df_ssds(df) for _, df in stopdf.groupby(g_cols)]
                ssd_list = [np.sort(issd).tolist() for issd in ssds]
            else:
                ssd_list = [['sacc'] for i in range(self.nrows)]
        self.make_idx_cols(ssd_list)
        masterDF_header = g_cols + self.p_cols + self.q_cols
        return masterDF_header

    def make_idx_cols(self, ssd_list=None):
        """ make idx-specific headers in event of missing data
        make all other headers if not done yet
        """
        if not hasattr(self, 'q_cols'):
            self.make_q_cols()
        if not hasattr(self, 'p_cols'):
            self.make_p_cols(ssd_list)
        if not hasattr(self, 'f_cols'):
            self.make_f_cols()
        if ssd_list:
            acc = [self.p_cols[0]]
            self.idx_cols = [acc + issd + self.q_cols for issd in ssd_list]
        else:
            self.idx_cols = [self.p_cols + self.q_cols]*self.nrows

    def make_q_cols(self):
        """ make header names for correct/error RT quants
        in observedDF, yhatDF, and wtsDF
        """
        cq = ['c' + str(int(n * 100)) for n in self.model.quantiles]
        eq = ['e' + str(int(n * 100)) for n in self.model.quantiles]
        self.q_cols = cq + eq

    def make_p_cols(self, ssd_list=None):
        """ make header names for response accuracy in observedDF,
        yhatDF, and wtsDF (including SSDs if stop model)
        """
        self.p_cols = ['acc']
        if ssd_list:
            ssd_unique = np.unique(np.hstack(ssd_list))
            self.p_cols = self.p_cols + ssd_unique.tolist()

    def make_f_cols(self):
        """ make header names for various fit statistics in fitDF
        (model parameters, goodness-of-fit measures, etc)
        """
        params = np.sort(list(self.inits))
        if not self.model.is_flat:
            dep_keys = list(self.model.pc_map)
            cond_param_names = listvalues(self.model.clmap)
            params = np.hstack([params, np.squeeze(cond_param_names)]).tolist()
            _ = [params.remove(pname) for pname in dep_keys]
        fit_cols = ['nfev', 'nvary', 'df', 'chi', 'rchi', 'logp', 'AIC', 'BIC', 'cnvrg']
        self.f_cols = np.hstack([['idx'], params, fit_cols]).tolist()

    def write_results(self, save_observed=False):
        """ Saves yhatDF and fitDF results to model output dir
        ::Arguments::
            save_observed (bool):
                if True will write observedDF & wtsDF to
                model output dir
        """
        fname = self.model.model_id
        if self.model.is_nested:
            fname='nested_models'
        make_fname = lambda savestr: '_'.join([fname, savestr+'.csv'])
        self.yhatDF.to_csv(make_fname('yhat'))
        self.fitDF.to_csv(make_fname('finfo'))
        if save_observed:
            self.observedDF.to_csv(make_fname('observed_data'))
            self.wtsDF.to_csv(make_fname('cost_weights'))

    def make_results_dir(self, custompath=None, get_path=True):
        """ make directory for writing model output and figures
        dir is named according to model_id, navigate to dir
        after ensuring it exists
        """
        savepath = os.path.expanduser('~')
        if custompath is not None:
            savepath = os.path.join(savepath, custompath)
        abspath = os.path.abspath(savepath)
        abs_savepath = os.path.join(abspath, self.model.model_id)
        if not os.path.isdir(abs_savepath):
            os.makedirs(abs_savepath)
        os.chdir(abs_savepath)
        if get_path:
            return abs_savepath



#!/usr/local/bin/env python
from __future__ import division
from future.utils import listvalues
from copy import deepcopy
import numpy as np
import pandas as pd
from numpy import array
from radd import theta
from radd.models import Simulator
from lmfit import minimize, fit_report
from scipy.optimize import basinhopping
from numpy.random import uniform

class BasinBounds(object):
    """ sets conditions for step acceptance during
    basinhopping optimization routine
    Arguments:
        xmin (list): lower boundaries for each parameter
        xmax (list): upper boundaries for each parameter
    """
    def __init__(self, xmin, xmax):
        self.xmin = np.array(xmin)
        self.xmax = np.array(xmax)

    def __call__(self, **kwargs):
        x = kwargs["x_new"]
        tmin = bool(np.all(x >= self.xmin))
        tmax = bool(np.all(x <= self.xmax))
        return tmin and tmax

class HopStep(object):
    """ scale stepsize of basinhopping optimization according to
    individual parameters ranges (smaller steps for more sensitive params).
    See theta.get_stepsize_scalars() for parameter <--> scalar mapping.

    Arguments:
        keys (list): list of parameter names
        stepsize (list): initial stepsize
    """
    def __init__(self, keys, nlevels=1, stepsize=0.05):
        self.stepsize_scalars = theta.get_stepsize_scalars(keys, nlevels)
        self.stepsize = stepsize
        self.np = self.stepsize_scalars.size

    def __call__(self, x):
        s = self.stepsize
        ss = self.stepsize_scalars
        x = np.array([x[i] + uniform(-ss[i]*s, ss[i]*s) for i in range(self.np)])
        return x

class Optimizer(object):
    """ Optimizer class acts as interface between Model and Simulator objects.
    Structures fitting routines so that Models are first optimized with the full set of
    parameters free, data collapsing across conditions.

    The fitted parameters are then used as the initial parameters for fitting conditional
    models with only a subset of parameters are left free to vary across levels of a given
    experimental conditions.

    Parameter dependencies are specified when initializing Model object via
    <depends_on> arg (i.e.{parameter: condition})

    Handles fitting routines for models of average, individual subject, and bootstrapped data
    """
    def __init__(self, fitparams=None, basinparams=None, inits=None, kind='xdpm', depends_on=None, pc_map=None):
        self.inits=inits
        self.fitparams = fitparams
        self.basinparams = basinparams
        self.kind = kind
        self.pc_map = pc_map
        self.pnames = ['a', 'tr', 'v', 'ssv', 'z', 'xb', 'si', 'sso']
        self.constants = deepcopy(['a', 'tr', 'v', 'xb'])

    def hop_around(self, inits, pbars=None):
        """ initialize model with niter randomly generated parameter sets
        and perform global minimization using basinhopping algorithm
        ::Arguments::
            p (dict):
                parameter dictionary
        ::Returns::
            parameter set with the best fit
        """
        callback = None
        xpopt, xfmin = [], []
        for i, p in enumerate(inits):
            if pbars is not None:
                pbars.update(name='glb_basin', i=i)
                callback = pbars.callback
            popt, fmin = self.run_basinhopping(p=p, callback=callback)
            xpopt.append(popt)
            xfmin.append(fmin)
        if self.fitparams['disp']:
            # report global minimum
            print("Finished Hopping Around:\nGLOBAL MIN = {:.9f}".format(np.min(xfmin)))
        # return parameters at the global basin
        return xpopt[np.argmin(xfmin)]

    def run_basinhopping(self, p, callback=None):
        """ uses fmin_tnc in combination with basinhopping to perform bounded global
         minimization of multivariate model
        ::Arguments::
            p (dict):
                parameter dictionary
            callback (function):
                callable function for displaying optimization progress
        """
        bp = self.basinparams
        nl = self.fitparams['nlevels']
        if nl==1:
            basin_keys = np.sort(list(p))
            xp = dict(deepcopy(p))
            basin_params = theta.scalarize_params(xp)
        else:
            basin_keys = np.sort(list(self.pc_map))
            basin_params = deepcopy(p)
        self.simulator.__prep_global__(basin_params=basin_params, basin_keys=basin_keys)
        # make list of init values for all pkeys included in fit
        x0 = np.hstack(np.hstack([basin_params[pk]*np.ones(nl) for pk in basin_keys]))
        # define parameter boundaries for all params in pc_map.keys()
        # to be used by basinhopping minimizer & tnc local optimizer
        xmin, xmax = theta.format_basinhopping_bounds(basin_keys, nlevels=nl, kind=self.kind)
        tncbounds = theta.format_local_bounds(xmin, xmax)
        tncopt = {'xtol': bp['tol'], 'ftol': bp['tol']}
        mkwargs = {"method": bp['method'], 'bounds': tncbounds, 'tol': bp['tol'], 'options': tncopt}
        # define custom take_step and accept_test functions
        accept_step = BasinBounds(xmin, xmax)
        custom_step = HopStep(basin_keys, nlevels=nl, stepsize=bp['stepsize'])
        # run basinhopping on simulator.basinhopping_minimizer func
        out = basinhopping(self.simulator.global_cost_fx, x0=x0, minimizer_kwargs=mkwargs, take_step=custom_step, accept_test=accept_step, T=bp['T'], stepsize=bp['stepsize'], niter_success=bp['nsuccess'], interval=bp['interval'], disp=bp['disp'], callback=callback)
        xopt = out.x
        funcmin = out.fun
        if nl > 1:
            xopt = [array([xopt]).reshape(len(basin_keys), nl).squeeze()]
        for i, k in enumerate(basin_keys):
            p[k] = xopt[i]
        return p, funcmin

    def gradient_descent(self, p=None):
        """ Optimizes parameters following specified parameter
        dependencies on task conditions (i.e. depends_on={param: cond})
        """
        fp = self.fitparams
        if p is None:
            p = dict(deepcopy(self.inits))
        optkws = {'disp': fp['disp'], 'xtol': fp['tol'], 'ftol': fp['tol'], 'maxfev': fp['maxfev']}
        # make lmfit Parameters object to keep track of
        # parameter names and dependencies during fir
        lmParams = theta.loadParameters(inits=p, pc_map=self.pc_map, is_flat=fp['flat'], kind=self.kind)
        lmMinimizer = minimize(self.simulator.cost_fx, lmParams, method=fp['method'], options=optkws)
        self.lmMinimizer = lmMinimizer
        #self.lmMinimizer = deepcopy(lmMinimizer)
        self.param_report = fit_report(lmMinimizer.params)
        # gen dict of lmfit optimized Parameters object
        finfo = pd.Series(lmMinimizer.params.valuesdict())
        p = finfo.to_dict()
        finfo['cnvrg'] = lmMinimizer.success
        finfo['nfev'] = lmMinimizer.nfev
        finfo['nvary'] = len(lmMinimizer.var_names)
        # get model-predicted yhat vector
        yhat = (lmMinimizer.residual / self.simulator.wts) + self.simulator.y
        # un-vectorize all parameters except conditionals
        popt = theta.scalarize_params(p, pc_map=self.pc_map, is_flat=fp['flat'])
        return finfo, popt, yhat



#!/usr/local/bin/env python
from __future__ import division
from copy import deepcopy
import numpy as np
from numpy import array
from numpy.random import sample as rs
from numpy import hstack as hs
from numpy import newaxis as na
from numpy import cumsum as csum
from scipy.stats.mstats import mquantiles as mq
from radd import theta

class Simulator(object):
    """ Core code for simulating models. All cond, trials, &
    timepoints are simulated simultaneously
    """
    def __init__(self, fitparams=None, pc_map=None, kind='xdpm', dt=.005, si=.01, learn=False, dynamic=False):
        self.learn = learn
        self.kind = kind
        self.ntime = 0
        self.dynamic = dynamic
        if 'ssd_info' in fitparams.keys():
            self.ssd_info = fitparams['ssd_info']
            self.include_ss=True
        if 'x' in self.kind:
            self.dynamic = True
        self.pc_map = pc_map
        self.fitparams = fitparams
        self.__update_steps__(dt=dt, si=si)
        self.__update__(fitparams=fitparams)

    def __update__(self, fitparams=None):
        """ update critical simulator parameters for each fit
        by providing an updated fitparams dictionary"""
        if fitparams is not None:
            self.fitparams = fitparams
        fp = deepcopy(self.fitparams)
        # non conditional parameters & meta-data
        self.pvc = ['a', 'tr', 'v', 'xb', 'ssv']
        self.tb = fp['tb']
        self.nconds = len(list(fp['clmap']))
        self.nlevels = fp['nlevels']
        self.ntot = fp['ntrials']
        self.quantiles = fp['quantiles']
        # set empirical data, wts vectors
        self.y = fp['y'].flatten()
        self.wts = fp['wts'].flatten()
        # include SSD's if stop-signal task
        if self.include_ss:
            self.ssd_info = fp['ssd_info']
        if self.nlevels>1:
            # remove any parameters free to vary across experimental conditions
            map((lambda pkey: self.pvc.remove(pkey)), list(self.pc_map))
        self.__update_rand_vectors__()
        self.__init_model_functions__()
        self.__init_analyze_functions__()

    def __update_steps__(self, dt=None, si=None, tb=None):
        """ update and store stepsize parameters
        dx: size of step up/down t-->t+dt
        dt: timestep of decision traces
        si: diffusion constant (noise)
        tb: time-boundary of trial
        """
        if tb is not None:
            self.tb = tb
        if si is not None:
            self.si = si
        if dt is not None:
            self.dt = dt
        self.dx = np.sqrt(self.si * self.dt)

    def __prep_global__(self, basin_params={}, basin_keys=[]):
        """ prepare simulator for global optimization using
        scipy.optimize.basinhopping algorithm
        """
        # set parameter names used to populate params dict
        self.basin_keys = basin_keys
        # set all constant parameters in basin_params object
        self.basin_params = basin_params
        self.chunk = lambda x, nl: [array(x[i:i+nl]) for i in range(0, len(x), nl)]

    def global_cost_fx(self, x):
        """ used specifically by fit.perform_basinhopping() for Model
        objects with multiopt attr (See __opt_routine__ and perform_basinhopping
        methods of Optimizer object)
        """
        p = dict(deepcopy(self.basin_params))
        # segment 'x' into equal len arrays (one array,
        # nlevels vals long per free parameter) in basin_keys
        px = self.chunk(x, self.nlevels)
        for i, pk in enumerate(self.basin_keys):
            p[pk] = px[i]
        # simulate using filled params dictionary
        yhat = self.sim_fx(p)
        # calculate and return cost error
        return np.sum((self.wts * (yhat - self.y))**2).astype(np.float32)

    def cost_fx(self, theta, sse=False):
        """ Main cost function used for fitting all models self.sim_fx
        determines which model is simulated (determined when Simulator
        is initiated)
        """
        if type(theta) == dict:
            p = dict(deepcopy(theta))
        else:
            p = theta.valuesdict()
        yhat = self.sim_fx(p, analyze=True)
        residuals = array(self.wts * (yhat - self.y))
        if sse:
            return np.sum(residuals**2).astype(np.float32)
        return residuals.astype(np.float32)

    def __init_model_functions__(self):
        """ initiates the simulation function used in
        optimization routine
        """
        if 'dpm' in self.kind:
            self.sim_fx = self.simulate_dpm
            self.analyze_fx = self.analyze_reactive
        elif 'pro' in self.kind:
            self.sim_fx = self.simulate_pro
            self.analyze_fx = self.analyze_proactive
        elif 'irace' in self.kind:
            self.sim_fx = self.simulate_irace
            self.analyze_fx = self.analyze_reactive
        # dynamic bias is hyperbolic cosine
        if self.dynamic:
            self.dynamics_fx = lambda p, t: np.cosh(p['xb'][:, na] * t)
        else:
            self.dynamics_fx = lambda p, t: np.ones((self.nlevels, len(t)))

    def __init_analyze_functions__(self):
        """ initiates the analysis function used in
        optimization routine to produce the yhat vector
        """
        prob = self.quantiles
        go_axis, ss_axis = 2, 3
        if self.learn:
            go_axis-=1; ss_axis-=1
        self.go_resp = lambda trace, upper: np.argmax((trace.T >= upper).T, axis=go_axis) * self.dt
        self.ss_resp_up = lambda trace, upper: np.argmax((trace.T >= upper).T, axis=ss_axis) * self.dt
        self.ss_resp_lo = lambda trace, x: np.argmax((trace.T <= 0).T, axis=ss_axis) * self.dt
        self.go_RT = lambda ontime, rbool: ontime[:, na] + (rbool*np.where(rbool==0., np.nan, 1))
        self.ss_RT = lambda ontime, rbool: ontime[:, :, na] + (rbool*np.where(rbool==0., np.nan, 1))
        self.RTQ = lambda zpd: map((lambda x: mq(x[0][x[0] < x[1]], prob)), zpd)
        if 'irace' in self.kind:
            self.ss_resp = self.ss_resp_up
        else:
            self.ss_resp = self.ss_resp_lo

    def vectorize_params(self, p):
        """ ensures that all parameters are converted to arrays before simulation. see
        doc strings for prepare_fit() method of Model class (in build.py) for details
        regarding pc_map and logic for fitting models with parameters that depend on
        experimental conditions
        ::Arguments::
            p (dict):
                dictionary with all model parameters as
                scalars/vectors/or both
        ::Returns::
            p (dict):
                dictionary with all parameters as vectors
        """
        nl_ones = np.ones(self.nlevels).astype(np.float32)
        if 'si' in list(p):
            self.dx = np.sqrt(p['si'] * self.dt)
        if 'xb' not in list(p):
            p['xb'] = 1.0
        if self.nlevels==1:
            p = theta.scalarize_params(p)
            return {pk:p[pk]*nl_ones for pk in list(p)}
        for pkey in self.pvc:
            p[pkey] = p[pkey] * nl_ones
        for pkey, pkc in self.pc_map.items():
            if pkc[0] not in list(p):
                p[pkey] = p[pkey] * nl_ones
            else:
                p[pkey] = array([p[pc] for pc in pkc]).astype(np.float32)
        return p

    def __update_rand_vectors__(self):
        """ update rvector (random_floats) for Go and Stop traces
        """
        nl, ntot, ntime = self.nlevels, self.ntot, self.ntime
        self.rvector = rs((nl, ntot, ntime))
        if self.include_ss:
            ssd, nssd, nss, nss_per, ssd_ix = self.ssd_info
            self.rvector_ss=self.rvector[:, :nss, :].reshape(nl, nssd, nss_per, ntime)

    def __update_trace_params__(self, p):
        """ update Pg (probability of DVg +dx) and Tg (n timepoints)
        for go process and get get dynamic bias signal if 'x' model
        """
        Pg = 0.5 * (1 + p['v'] * self.dx / self.si)
        Tg = np.ceil((self.tb - p['tr']) / self.dt).astype(int)
        self.ntime_new = Tg.max()
        out = [Pg]
        if self.include_ss:
            out.extend(self.__update_ss_trace_params__(p, Tg))
        if self.ntime_new > self.ntime:
            self.ntime = self.ntime_new
            self.__update_rand_vectors__()
        self.xtb = self.dynamics_fx(p, csum([self.dt] * self.ntime))
        return out

    def __update_ss_trace_params__(self, p, Tg, sso=0):
        """ update Ps (probability of DVs +dx) and Ts (n timepoints)
        for condition and each SSD of stop process
        """
        ssd = self.ssd_info[0]
        if 'sso' in list(p):
            sso = p['sso']
        Ps = 0.5 * (1 + p['ssv'] * self.dx / self.si)
        Ts = np.ceil((self.tb - (ssd + sso)) / self.dt).astype(int)
        ss_on = 0
        if 'dpm' in self.kind:
            ss_on = np.where(Ts<Tg[:, na], Tg[:, na]-Ts, ss_on)
        self.ntime_new = np.max([self.ntime_new, Ts.max()])
        return [Ps, ss_on]

    def simulate_dpm(self, p, analyze=True):
        """ Simulate the dependent process model (DPM)
        ::Arguments::
            p (dict):
                parameter dictionary. values
            analyze (bool <True>):
                if True (default) return rt and accuracy information
                else, return Go and Stop decision traces.
        ::Returns::
            yhat of cost vector (ndarray)
            or list of decision traces (list of ndarrays)
        """
        p = self.vectorize_params(p)
        nl, ntot, dx = self.nlevels, self.ntot, self.dx
        ssd, nssd, nss, nss_per, ssd_ix = self.ssd_info
        Pg, Ps, ss_on = self.__update_trace_params__(p)
        # generate Go traces (nlevels, ntrials, ntimepoints)
        DVg = self.xtb[:,na] * csum(np.where(self.rvector.T < Pg, dx, -dx).T, axis=2)
        ssDVg = DVg[:, :nss, :].reshape(nl, nssd, nss_per, DVg.shape[-1])
        # use array-indexing to initialize SS at DVg[:nlevels, :ssd, :trials, t=SSD]
        ssBase = ssDVg[np.arange(nl)[:,na], ssd_ix, :, ss_on][:,:,:,na]
        # add ssBaseline to SS traces (nlevels, nSSD, ntrials_perssd, ntimepoints)
        DVs = ssBase + csum(np.where(self.rvector_ss.T < Ps, dx, -dx).T, axis=3)
        if analyze:
            return self.analyze_fx(DVg, DVs, p)
        return [DVg, DVs]

    def simulate_irace(self, p, analyze=True):
        """ simulate the independent race model
        (see simulate_dpm() for I/O details)
        """
        p = self.vectorize_params(p)
        nl, ntot, dx = self.nlevels, self.ntot, self.dx
        ssd, nssd, nss, nss_per, ssd_ix = self.ssd_info
        Pg, Ps, ss_on = self.__update_trace_params__(p)
        # generate Go traces (nlevels, ntrials, ntimepoints)
        DVg = self.xtb[:,na] * csum(np.where(self.rvector.T < Pg, dx, -dx).T, axis=2)
        # generate SS traces (nlevels, nSSD, ntrials_perssd, ntimepoints)
        DVs = csum(np.where(self.rvector_ss.T < Ps, dx, -dx).T, axis=3)
        if analyze:
            return self.analyze_fx(DVg, DVs, p)
        return [DVg, DVs]

    def simulate_pro(self, p, analyze=True):
        """ Simulate the proactive competition model
        (see simulate_dpm() for I/O details)
        """
        p = self.vectorize_params(p)
        nl, ntot, dx = self.nlevels, self.ntot, self.dx
        Pg = self.__update_trace_params__(p)
        # generate Go traces (nlevels, ntrials, ntimepoints)
        DVg = self.xtb[:,na] * csum(np.where(self.rvector.T < Pg, dx, -dx).T, axis=2)
        if analyze:
            return self.analyze_fx(DVg, p)
        return DVg

    def analyze_reactive(self, DVg, DVs, p):
        """ get rt and accuracy of go and stop process for simulated
        conditions generated from simulate_dpm
        """
        ssd, nssd, nss, nss_per, ssd_ix = self.ssd_info
        nl, ntot = self.nlevels, self.ntot
        gdec = self.go_resp(DVg, p['a'])
        # if dpm, simply ss_resp() uses 0
        # as boundary simply ignores sec. arg
        sdec = self.ss_resp(DVs, p['a'])
        gort = self.go_RT(p['tr'], gdec)
        ssrt = self.ss_RT(ssd, sdec)
        ert = gort[:, :nss].reshape(nl, nssd, nss_per)
        eq = self.RTQ(zip(ert, ssrt))
        gq = self.RTQ(zip(gort, [self.tb] * nl))
        gacc = np.nanmean(np.where(gort < self.tb, 1, 0), axis=1)
        sacc = np.where(ert < ssrt, 0, 1).mean(axis=2)
        return hs([hs([i[ii] for i in [gacc, sacc, gq, eq]]) for ii in range(nl)])

    def analyze_proactive(self, DVg, p):
        """ get proactive rt and accuracy of go process for simulated
        conditions generated from simulate_pro
        """
        nl, ntot = self.nlevels, self.ntot
        gdec = self.go_resp(DVg, p['a'])
        gort = self.go_RT(p['tr'], gdec)
        gq = self.RTQ(zip(gort, [self.tb] * nl))
        # Get response and stop accuracy information
        gacc = 1 - np.mean(np.where(gort < self.tb, 1, 0), axis=1)
        return hs([gacc, gq])

    def simulate_rldpm(self, p, analyze=True):
        """ Simulate the dependent process model (DPM)
        with learning
        """
        p = self.vectorize_params(p)
        Pg, xtb, Ps, ss_on = self.__update_go_process__(p)
        nl, ntot, dx = self.nlevels, self.ntot, self.dx
        ssd, nssd, nss, nss_per, ssd_ix = self.ssd_info

        for trial in range(ntot):
            DVg = xtb[:, na] * csum(np.where(self.rvector[:, trial, :].T < Pg, dx, -dx).T, axis=2)
            # INITIALIZE DVs FROM DVg(t=SSD)
            if trial%2:
                DVg[:, ]
                ssBase = ssDVg[np.arange(nl)[:,na], ssd_ix, :, ss_on][:,:,:,na]
                DVs = ssBase + csum(np.where(self.rvector_ss.T < Ps, dx, -dx).T, axis=3)
                #ssBase = DVg[np.arange(nl)[:,na], ssd_ix, :, ss_on][:,:,:,na]
                #DVs = init_ss[:,:,na] + csum(np.where(rs((nss, Ts.max()))<Ps, dx, -dx), axis=1)
                DVs = csum(np.where(rs((nl, Ts.max()))<Ps, dx, -dx), axis=1)
        if analyze:
            return self.analyze_fx(DVg, DVs, p)
        return [DVg, DVs]



#!/usr/local/bin/env python
from __future__ import division
import sys
from copy import deepcopy
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from radd.tools import colors, analyze
import warnings
warnings.simplefilter('ignore', np.RankWarning)

cdict = colors.get_cpals('all')
rpal = cdict['rpal']
bpal = cdict['bpal']
gpal = cdict['gpal']
ppal = cdict['ppal']
heat = cdict['heat']
cool = cdict['cool']
slate = cdict['slate']

def plot_model_fits(y, yhat, fitparams, err=None, palettes=[gpal, bpal], save=False, cdf=True, bw=.01, sameaxis=False):
    """ main plotting function for displaying model fit predictions over data
    """
    sns.set(style='darkgrid', rc={'figure.facecolor':'white'}, font_scale=1.5)
    nlevels = fitparams['nlevels']
    ssd, nssd, nss, nss_per, ssd_ix = fitparams.ssd_info
    quantiles = fitparams.quantiles
    if len(palettes) != nlevels:
        palettes = colors.get_cpals(aslist=True)[:nlevels]
    if sameaxis or nlevels==1:
        f, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(14, 5), sharey=True)
        axes = np.asarray([ax1, ax2, ax3])
    else:
        f, axes = plt.subplots(nlevels, 3, figsize=(14, 4.5*nlevels), sharey=True)
    if nssd==1:
        plot_acc = plot_accuracy
    else:
        plot_acc = scurves
    clrs = [pal(2) for pal in palettes]
    lbls = get_plot_labels(fitparams)
    y_dat = unpack_vector(y, fitparams, bw=bw)
    y_kde = unpack_vector(y, fitparams, bw=bw, kde=True)
    yhat_dat = unpack_vector(yhat, fitparams, bw=bw)
    if err is not None:
        y_err = unpack_vector(err, fitparams, bw=bw)
        sc_err = [e[0] for e in y_err]
        qp_err = [[e[1], e[2]] for e in y_err]
    else:
        sc_err, qp_err = [[None]*nlevels]*2
    for i in range(nlevels):
        if not sameaxis:
            ax1, ax2, ax3 = axes[i]
        accdata = [y_dat[i][0], yhat_dat[i][0]]
        qpdata = [y_dat[i], yhat_dat[i]]
        plot_acc(accdata, err=sc_err[i], ssd=ssd[i], colors=clrs[i], labels=lbls[i], ax=ax1, ssdlabels=sameaxis)
        plot_quantiles(qpdata, err=qp_err[i], quantiles=quantiles, colors=clrs[i], axes=[ax2,ax3], kde=y_kde[i], bw=bw)
    axes = format_axes(axes)
    if save:
        plt.savefig(fitparams['model_id']+'.png', dpi=600)
    if fitparams['fit_on']=='subjects' and save:
        plt.close('all')

def scurves(data, err=None, colors=None, labels=None, ssd=None, ax=None, get_pse=False, pretune_fx='interpolate', **kwargs):
    """ plotting function for displaying model-predicted
    stop curve (across SSDs) over empirical estimates
    """
    if err is None:
        err = [np.zeros(len(dat)) for dat in data]
    if ax is None:
        f, ax = plt.subplots(1, figsize=(5.5, 6))
    if colors is None:
        colors = slate(len(data))
    if labels is None:
        labels = [''] * len(data)
    if ssd is not None:
        ssds = ssd.squeeze()*1e3
        x = np.array([np.float(d) for d in ssds])
        xtls = ["{:.0f}".format(d) for d in ssds]
    else:
        x = np.arange(len(data[0]), dtype='float')*50
        xtls = [str(xx) for xx in x]
    pse=[]
    for i, yi in enumerate(data):
        if pretune_fx=='interpolate':
            xsim, ysim = analyze.scurve_interpolate(x, yi)
        elif pretune_fx=='polynomial':
            xsim, ysim = analyze.scurve_poly_fit(x, yi)
        xsim, ysim = analyze.fit_sigmoid(xsim, ysim)
        idx = (np.abs(ysim - .5)).argmin()
        pse.append(xsim[idx])
        if not i%2:
            ax.errorbar(x, yi, yerr=err, lw=0., elinewidth=1.5, color=colors[i], marker='o', ms=5)
            ax.plot(xsim, ysim, lw=2, color=colors[i], label=labels[i])
            continue
        ax.plot(x, yi, lw=0., color=colors[i], marker='o', ms=10, mfc="none", mec=colors[i], mew=1.7, label=labels[i])
    xxlim = (xsim[0], xsim[-1])
    ytls = np.arange(0, 1.2, .2).astype(np.int)
    plt.setp(ax, xticks=x, xlim=xxlim, xticklabels=xtls, ylim=(-.01, 1.05), yticks=ytls, yticklabels=ytls)
    ax.set_ylabel('Percent Correct'); ax.legend(loc=0)
    ax.legend(loc=0)
    plt.tight_layout()
    sns.despine()
    if get_pse:
        return pse

def plot_quantiles(data, err=None, quantiles=None, axes=None, colors=None, labels=None, kde=None, bw=.008):
    """ plotting function for displaying model-predicted
    quantile-probabilities over empirical estimates
    """
    y_data, yhat_data = data
    c, c_hat = colors
    if axes is not None:
        axc, axe = axes
    else:
        f, (axc, axe) = plt.subplots(1, 2, figsize=(10, 4))
    qc, qc_hat = y_data[1], yhat_data[1]
    qe, qe_hat = y_data[2], yhat_data[2]
    if quantiles is None:
        quantiles = np.linspace(.1, .9, qc.size)
    if err is not None:
        qc_err, qe_err = err
    else:
        qc_err, qe_err = [np.zeros(len(qc))]*2
    if kde is not None:
        qc_kde, qe_kde = kde[1], kde[2]
        sns.kdeplot(qc_kde, cumulative=1, color=c, ax=axc, linewidth=2, linestyle='-', bw=bw)
        sns.kdeplot(qe_kde, cumulative=1, color=c, ax=axe, linewidth=2, linestyle='-', bw=bw)
    axc.errorbar(qc, quantiles, xerr=qc_err, color=c, linewidth=0, elinewidth=1.5, marker='o', ms=5, label=labels)
    axe.errorbar(qe, quantiles, xerr=qe_err, color=c, linewidth=0, elinewidth=1.5, marker='o', ms=5, label=labels)
    axc.plot(qc_hat, quantiles, mec=c_hat, linewidth=0, marker='o', ms=10, mfc='none', mew=1.7, label=labels)
    axe.plot(qe_hat, quantiles, mec=c_hat, linewidth=0, marker='o', ms=10, mfc='none', mew=1.7, label=labels)

def plot_accuracy(data=[], err=None, ssd=None, ax=None, linestyles=None, colors=None, labels=None, xtls=['Go', 'Stop'], ssdlabels=False, **kwargs):
    """ plotting function for displaying model-predicted
    stop probability (at mean SSD) on top of empirical estimates
    (used when data is collected using tracking procedure to est. SSRT)
    """
    if err is None:
        err = [np.zeros(len(dat)) for dat in data]
    if ax is None:
        f, ax = plt.subplots(1, figsize=(5.5, 6))
    if colors is None:
        colors = slate(len(data))
    if labels is None:
        labels = [''] * len(data)
    if linestyles is None:
        linestyles = ['-', '--'] * len(data)
    x = np.arange(len(data[0]), dtype='float')
    if ssd is not None:
        ssdstr = "SSD={:.2f}".format(ssd.mean(axis=0)*1e3)
        if ssdlabels:
            labels[0] = ' '.join([labels[0], ssdstr])
            xtls = ['Go', 'Stop']
        else:
            xtls = ['Go', ssdstr]
    for i, yi in enumerate(data):
        if not i%2:
            ax.errorbar(x, yi, yerr=err[i], linestyle=linestyles[i], lw=1.5, elinewidth=1.5, color=colors[i], label=labels[i], marker='o', ms=5)
            continue
        xjit = x + (i*.02)
        ax.plot(xjit, yi, lw=0, marker='o', ms=10, mfc='none', mec=colors[i], mew=1.7, label=labels[i])

    plt.setp(ax, xticks=x, xlim=(-0.25, 1.25), xticklabels=xtls, ylim=(-.01, 1.05), yticks=np.arange(0, 1.2, .2))
    ax.set_ylabel('Percent Correct'); ax.legend(loc=0)
    sns.despine()
    plt.tight_layout()

def unpack_vector(vector, fitparams, kde=False, bw=.01):
    nlevels = fitparams['nlevels']
    nquant = fitparams['quantiles'].size
    nssd = fitparams['ssd_info'][1]
    vector = vector.reshape(nlevels, int(vector.size/nlevels))
    unpacked = []
    for v in vector:
        if nssd>1:
            # get accuracy at each SSD
            presp = v[1:1+nssd]
        else:
            # get go, stop accuracy
            presp = v[:2]
        quant = v[-nquant*2:]
        quant_cor = quant[:nquant]
        quant_err = quant[-nquant:]
        if kde:
            quant_cor = analyze.kde_fit_quantiles([quant_cor], bw=bw)
            quant_err = analyze.kde_fit_quantiles([quant_err], bw=bw)
        unpacked.append([presp, quant_cor, quant_err])
    return unpacked

def format_axes(axes):
    q_axes = []
    for ax in axes.flatten():
        if not ax.is_first_col():
            q_axes.append(ax)
    yylim=(-.01, 1.05); yyticks=np.arange(0, 1.2, .2)
    axesqp = [np.hstack([l.get_xdata() for l in ax.lines]) for ax in q_axes]
    xxticks = [[qdata.min(), qdata.mean(), qdata.max()] for qdata in axesqp]
    xxlim = [(xxt[0]-.008, xxt[-1]+.008) for xxt in xxticks]
    xxtls = [["{:.2f}".format(xtl) for xtl in xxt] for xxt in xxticks]
    for ax in axes.flatten():
        plt.setp(ax, ylim=yylim, yticks=yyticks)
        if not ax.is_first_col():
            plt.setp(ax, yticklabels=yyticks)
        if not ax.is_first_col():
            xl, xt, xtl = xxlim[0], xxticks[0], xxtls[0]
            if ax.is_last_col():
                xl, xt, xtl = xxlim[-1], xxticks[-1], xxtls[-1]
            plt.setp(ax, xlim=xl, xticks=xt, xticklabels=xtl)
            if ax.is_last_row():
                plt.setp(ax, xlabel='RT (s)')
            else:
                plt.setp(ax, xlabel='')
        if ax.is_first_col() and ax.is_last_row():
            ax.set_xlabel('SSD (ms)')
    axes = axes.flatten()
    axes[0].set_title('Accuracy', position=(.5, 1.02))
    axes[1].set_title('Correct Quant-Prob', position=(.5, 1.02))
    axes[2].set_title('Error Quant-Prob', position=(.5, 1.02))
    sns.despine()
    plt.tight_layout()
    return axes

def compare_nested_models(fitDF, model_ids, plot_stats=True):
    gof = {}
    fitDF = fitDF[fitDF.idx.isin(model_ids)]
    # print GOF stats for both models
    for i, p in enumerate(model_ids):
        gof[p] = fitDF.loc[i, ['AIC', 'BIC']]
        print('{} GOF:\nAIC = {}\nBIC = {}\n'.format(model_ids[i], *gof[p]))
    # Which model provides a better fit to the data?
    aicwinner = model_ids[np.argmin([gof[mid][0] for mid in model_ids])]
    bicwinner = model_ids[np.argmin([gof[mid][1] for mid in model_ids])]
    print('AIC likes {} model'.format(aicwinner))
    print('BIC likes {} model'.format(bicwinner))
    plot_model_gof(gof, aicwinner)

def plot_model_gof(gof_dict, aicwinner):
    sns.set(style='darkgrid', rc={'figure.facecolor':'white'}, font_scale=1.5)
    nmodels = len(list(gof_dict))
    c1 = ["#4168B7", "#d62c1a"]
    l1, l2 = None, None
    f, ax = plt.subplots(1, figsize=(8,6))
    x = np.arange(1, nmodels*2, 2)
    vals = np.hstack(gof_dict.values()).astype(float)
    yylim = (vals.max()*1.05, vals.min()*.95)
    xtls = []
    for i, (m_id, gof) in enumerate(gof_dict.items()):
        if i == len(x)-1:
            l1, l2 = 'AIC', 'BIC'
        a=.4
        if m_id==aicwinner:
            a=1
        ax.plot(x[i]-.2, gof['AIC'], color=c1[0], marker='o', ms=10, label=l1, alpha=a)
        ax.plot(x[i]+.2, gof['BIC'], color=c1[1], marker='o', ms=10, label=l2, alpha=a)
        xtls.append(m_id)
    plt.setp(ax, xticks=x, xticklabels=xtls, ylim=yylim, xlim=(0, nmodels*2))
    ax.xaxis.tick_top()
    sns.despine()
    ax.legend(loc=0)

def get_plot_labels(fitparams):
    lbls = np.hstack(fitparams['clmap'].values())
    labels = [[lbl + dtype for dtype in [' data', ' model']] for lbl in lbls]
    return labels



#!/usr/local/bin/env python
from __future__ import division
from future.utils import listvalues
from copy import deepcopy
import os
import numpy as np
import pandas as pd
from numpy import array
from scipy.stats.mstats import mquantiles as mq
from lmfit import fit_report
from radd.tools import messages
from radd import theta, vis

class RADDCore(object):
    """ Parent class for constructing attributes and methods used by
    of Model objects. Not meant to be used directly.

    Contains methods for building dataframes, generating observed data vectors
    that are entered into cost function during fitting as well as calculating
    summary measures and weight matrix for weighting residuals during optimization.
    """
    def __init__(self, data=None, kind='xdpm', inits=None, fit_on='average', depends_on={'all':'flat'}, quantiles=np.arange(.1, 1.,.1), ssd_method=None, weighted=True, verbose=False, custompath=None, nested_models=None):
        self.verbose = verbose
        self.kind = kind
        self.fit_on = fit_on
        self.ssd_method = ssd_method
        self.weighted = weighted
        self.quantiles = quantiles
        self.tb = data[data.response == 1].rt.max()
        self.idx = list(data.idx.unique())
        self.nidx = len(self.idx)
        self.depends_on = depends_on
        self.conds = np.unique(listvalues(depends_on)).tolist()
        if 'flat' in self.conds:
            self.is_flat = True
            data = data.copy()
            data['flat'] = 'flat'
        else:
            self.is_flat = False
        self.nconds = len(self.conds)
        self.clmap = {c: np.sort(data[c].unique()) for c in self.conds}
        self.nlevels = np.sum([len(lvls) for lvls in listvalues(self.clmap)])
        self.groups = np.hstack([['idx'], self.conds]).tolist()
        self.data = data
        self.inits = inits
        self.__prepare_fit__()
        self.finished_sampling = False
        self.track_subjects = False
        self.track_basins = False
        self.is_nested = False

    def __prepare_fit__(self):
        """ model setup and initiates dataframes. Automatically run when Model object is initialized
        *pc_map is a dict containing parameter names as keys with values
                corresponding to the names given to that parameter in Parameters object
                (see optmize.Optimizer).
        *Parameters (p[pkey]=pval) that are constant across conditions are broadcast as [pval]*n.
                Conditional parameters are treated as arrays with distinct values [V1, V2...Vn], one for
                each condition.
        pc_map (dict): see bound __format_pcmap__ method
        """
        from radd.optimize import Optimizer
        from radd.models import Simulator
        # initial parameters
        if self.inits is None:
            self.__get_default_inits__()
        # pc_map (see docstrings)
        self.__format_pcmap__()
        # create model_id string for naming output
        self.generate_model_id()
        # initialize DataHandler & generate I/O dataframes
        self.__make_dataframes__()
        # set fit parameters with default values
        self.set_fitparams()
        # set basinhopping parameters with default values
        self.set_basinparams()
        # initialize optimizer object for controlling fit routines
        # (updated with fitparams/basinparams whenever params are set)
        self.opt = Optimizer(fitparams=self.fitparams, basinparams=self.basinparams, kind=self.kind, inits=self.inits, depends_on=self.depends_on, pc_map=self.pc_map)
        # initialize model simulator, mainly accessed by the model optimizer object
        self.opt.simulator = Simulator(fitparams=self.fitparams, kind=self.kind, pc_map=self.pc_map)
        if self.verbose:
            self.is_prepared = messages.saygo(depends_on=self.depends_on, cond_map=self.clmap, kind=self.kind, fit_on=self.fit_on)
        else:
            self.prepared = True

    def __make_dataframes__(self):
        """ wrapper for dfhandler.DataHandler.make_dataframes
        """
        from radd.dfhandler import DataHandler
        # initialize dataframe handler
        self.handler = DataHandler(self)
        # make dataframes
        self.handler.make_dataframes()
        # Group dataframe (nsubjects*nconds*nlevels x ndatapoints)
        self.observedDF = self.handler.observedDF.copy()
        # list (nsubjects long) of data arrays (nconds*nlevels x ndatapoints) to fit
        self.observed = self.handler.observed
        # list of flattened data arrays (averaged across conditions)
        self.observed_flat = self.handler.observed_flat
        # dataframe with same dim as observeddf for storing model predictions
        self.yhatDF = self.handler.yhatDF
        # dataframe with same dim as observeddf for storing fit info
        self.fitDF = self.handler.fitDF
        # dataframe containing cost_function wts (see dfhandler docs)
        self.wtsDF = self.handler.wtsDF
        # list of arrays containing conditional costfx weights
        self.cond_wts = self.handler.cond_wts
        # list of arrays containing flat costfx weights
        self.flat_wts = self.handler.flat_wts
        # define iterables containing fit y & wts for each fit
        self.iter_flat = zip(self.observed_flat, self.flat_wts)
        self.iter_cond = zip(self.observed, self.cond_wts)

    def set_fitparams(self, **kwargs):
        """ dictionary of fit parameters, passed to Optimizer/Simulator objects
        """
        if not hasattr(self, 'fitparams'):
            # initialize with default values and first arrays in observed_flat, flat_wts
            self.fitparams = {'idx':0, 'y': self.observed_flat[0], 'wts': self.flat_wts[0],
                'ntrials': 20000, 'tol': 1.e-30, 'method': 'nelder', 'maxfev': 2000,
                'tb': self.tb, 'nlevels': 1, 'fit_on': self.fit_on, 'kind': self.kind,
                'clmap': self.clmap, 'quantiles': self.quantiles, 'model_id': self.model_id,
                'depends_on': self.depends_on, 'flat': True, 'disp':True}
            self.fitparams = pd.Series(self.fitparams)
        else:
            # fill with kwargs (i.e. y, wts, idx, etc) for the upcoming fit
            for kw_arg, kw_val in kwargs.items():
                self.fitparams[kw_arg] = kw_val
        if hasattr(self, 'ssd'):
            self.__set_ssd_info__()
        if self.fitparams.quantiles.size != self.quantiles.size:
            self.quantiles = self.fitparams.quantiles
            self.__make_dataframes__()
            self.fitparams['y'] = self.observed_flat[self.fitparams.idx]
            self.fitparams['wts'] = self.flat_wts[self.fitparams.idx]
        if list(self.depends_on)[0]!=list(self.fitparams.depends_on)[0]:
            self.depends_on = self.fitparams.depends_on
            self.__format_pcmap__()
            self.fitparams['model_id'] = self.generate_model_id(get_id=True)
            self.is_flat = False
        if hasattr(self, 'opt'):
            self.opt.fitparams = self.fitparams
            self.opt.pc_map = self.pc_map
            self.opt.simulator.pc_map = self.pc_map
            self.opt.simulator.__update__(fitparams=self.opt.fitparams)

    def set_basinparams(self, **kwargs):
        """ dictionary of global fit parameters, passed to Optimizer/Simulator objects
        """
        if not hasattr(self, 'basinparams'):
            self.basinparams =  {'ninits': 5, 'nsamples': 5000, 'interval': 10, 'T': 1.,
            'disp': False, 'stepsize': .05, 'nsuccess': 40, 'tol': 1.e-20, 'method': 'TNC',
            'init_sample_method': 'best'}
        else:
            # fill with kwargs for the upcoming fit
            for kw_arg, kw_val in kwargs.items():
                self.basinparams[kw_arg] = kw_val
            self.opt.basinparams = self.basinparams

    def __set_ssd_info__(self):
        """ set ssd_info for upcoming fit and store in fitparams dict
        """
        if self.fit_on=='average':
            ssd = np.array(self.ssd).mean(axis=0)
        else:
            # get ssd vector for fit number idx
            ssd = self.ssd[self.fitparams.idx]
        if self.fitparams.nlevels==1:
            # single vector (nlevels=1), don't squeeze
            ssd = np.mean(ssd, axis=0, keepdims=True)
        nssd = ssd.shape[-1]
        nss = int((.5 * self.fitparams.ntrials))
        nss_per_ssd = int(nss/nssd)
        ssd_ix = np.arange(nssd) * np.ones((ssd.shape[0], ssd.shape[-1])).astype(np.int)
        # store all ssd_info in fitparams, accessed by Simulator
        self.fitparams['ssd_info'] = [ssd, nssd, nss, nss_per_ssd, ssd_ix]

    def __format_pcmap__(self):
        """ dict used by Simulator to extract conditional parameter values by name
        from lmfit Parameters object
            |<--- PARAMETERS OBJECT [LMFIT] <------- [IN]
            |---> p = {'v_bsl': V1, 'v_pnl': V2...} --->|
            |<--- pc_map = {'v':['v_bsl', 'v_pnl']} <---|
            |---> p['v'] = array([V1, V2]) -------> [OUT]
        """
        pc_map = {}
        if not self.is_flat:
            for p, cond in self.depends_on.items():
                levels = np.sort(self.data[cond].unique())
                pc_map[p] = ['_'.join([p, lvl]) for lvl in levels]
        self.pc_map = pc_map
        if hasattr(self, 'opt'):
            self.handler.pc_map = self.pc_map
            self.opt.pc_map = self.pc_map
            self.opt.fitparams = self.fitparams
            self.opt.simulator.pc_map = self.pc_map
            self.opt.simulator.__update__(fitparams=self.opt.fitparams)

    def sample_param_sets(self, pkeys=None, nsamples=None):
        """ sample *nsamples* (default=5000, see set_fitparams) different
        parameter sets (param_sets) and get model yhat for each set (param_yhats)
        """
        if pkeys is None:
            pkeys = np.sort(list(self.inits))
        if nsamples is None:
            nsamples = self.basinparams['nsamples']
        self.param_sets = theta.random_inits(pkeys, ninits=nsamples, kind=self.kind, as_list_of_dicts=True)
        self.param_yhats = [self.opt.simulator.sim_fx(params_i) for params_i in self.param_sets]
        self.finished_sampling = True

    def filter_param_sets(self):
        """ sample *nsamples* (default=5000, see set_fitparams) different
        parameter sets (param_sets) and get model yhat for each set (param_yhats)
        """
        if not hasattr(self, 'param_sets'):
            self.sample_param_sets()
        nkeep = self.basinparams['ninits']
        keep_method = self.basinparams['init_sample_method']
        inits_list, globalmin = theta.filter_param_sets(self.param_sets, self.param_yhats, self.fitparams, nkeep=nkeep, keep_method=keep_method)
        return inits_list, globalmin

    def fill_yhatDF(self, yhat=None, fitparams=None):
        """ wrapper for filling & updating model yhatDF
        """
        if yhat is None:
            yhat = self.yhat
        if fitparams is None:
            fitparams = self.fitparams
        self.handler.fill_yhatDF(data=yhat, fitparams=fitparams)
        self.yhatDF = self.handler.yhatDF.copy()

    def fill_fitDF(self, finfo=None, fitparams=None):
        """ wrapper for filling & updating model fitDF
        """
        if finfo is None:
            finfo = self.finfo
        if fitparams is None:
            fitparams = self.fitparams
        self.handler.fill_fitDF(data=finfo, fitparams=fitparams)
        self.fitDF = self.handler.fitDF.copy()

    def write_results(self, save_observed=False):
        """ wrapper for dfhandler.write_results saves yhatDF and fitDF
        results to model output dir
        ::Arguments::
            save_observed (bool):
                if True will write observedDF & wtsDF to
                model output dir
        """
        self.handler.write_results(save_observed)

    def plot_model_fits(self, y=None, yhat=None, fitparams=None, kde=True, err=None, save=False, bw=.008, sameaxis=False):
        """ wrapper for radd.tools.vis.plot_model_fits """
        from radd import vis
        if fitparams is None:
            fitparams=self.fitparams
        if y is None:
            y = fitparams['y']
        if yhat is None:
            if hasattr(self, 'yhat'):
                yhat = self.yhat
            else:
                yhat = deepcopy(y)
                print("model is unoptimized, no yhat provided")
        if self.fit_on=='average' and err is None:
            err = self.handler.observed_err
        vis.plot_model_fits(y, yhat, fitparams, err=err, save=save, bw=bw, sameaxis=sameaxis)

    def log_fit_info(self, finfo, popt, fitparams):
        """ write meta-information about latest fit
        to logfile (.txt) in working directory
        """
        fp = dict(deepcopy(fitparams))
        # lmfit-structured fit_report to write in log file
        param_report = self.opt.param_report
        # log all fit and meta information in working directory
        messages.logger(param_report, finfo=finfo, popt=popt, fitparams=fp, kind=self.kind)

    def generate_model_id(self, get_id=False, appendstr=None):
        """ generate an identifying string with model information.
        used for reading and writing model output
        """
        model_id = list(self.depends_on)
        if 'all' in model_id:
            model_id = ['flat']
        model_id.append(self.fit_on)
        model_id.insert(0, self.kind)
        if appendstr is not None:
            model_id.append(appendstr)
        self.model_id = '_'.join(model_id)
        if get_id:
            return self.model_id

    def make_progress_bars(self, models=None):
        """ initialize progress bars to track fit progress (subject fits,
        init optimization, etc)
        """
        from radd.tools.utils import NestedProgress
        n = self.basinparams['ninits']
        self.pbars = NestedProgress(name='glb_basin', n=n, title='Global Basin', color='green')
        self.pbars.add_bar(name='lcl_basin', bartype='infobar', title='Current Basin', color='red')
        if self.is_nested:
            self.pbars.add_bar(name='models', n=self.nmodels, title='Nested Fits', color='blue')
        self.track_basins=True
        if self.fit_on=='subjects':
            self.track_subjects = True
            self.pbars.add_bar(name='idx', n=self.nidx, title='Subject Fits', color='blue')
        self.fitparams['disp']=False
        self.basinparams['disp']=False

    def __remove_outliers__(self, sd=1.5, verbose=False):
        """ remove slow rts (>sd above mean) from main data DF
        """
        from radd.tools.analyze import remove_outliers
        self.data = analyze.remove_outliers(self.data.copy(), sd=sd, verbose=verbose)

    def __get_default_inits__(self):
        """ if inits not provided by user, initialize with default values
        see tools.theta.get_default_inits
        """
        self.inits = theta.get_default_inits(kind=self.kind, depends_on=self.depends_on)

    def __check_inits__(self, inits):
        """ ensure inits dict is appropriate for Model kind
        see tools.theta.check_inits
        """
        inits = dict(deepcopy(inits))
        checked = theta.check_inits(inits=inits, depends_on=self.depends_on, kind=self.kind)
        return checked



#!usr/bin/env python
from __future__ import division
import pandas as pd
import numpy as np
import os
from numpy import array
from sklearn.neighbors import KernelDensity
from scipy.stats.mstats import mquantiles as mq
from scipy import optimize
import functools
from scipy.interpolate import interp1d

def remove_outliers(data, sd=1.5, verbose=False):
    df = data.copy()
    ssdf = df[df.response == 0]
    godf = df[df.response == 1]
    bound = godf.rt.std() * sd
    rmslow = godf[godf['rt'] < (godf.rt.mean() + bound)]
    clean_go = rmslow[rmslow['rt'] > (godf.rt.mean() - bound)]
    clean = pd.concat([clean_go, ssdf])
    if verbose:
        pct_removed = len(clean) * 1. / len(df)
        print("len(df): %i\nbound: %s \nlen(cleaned): %i\npercent removed: %.5f" % (len(df), str(bound), len(clean), pct_removed))
    return clean

def ensure_numerical_wts(wts, flat_wts):
    # test inf
    wts[np.isinf(wts)] = np.median(wts[~np.isinf(wts)])
    flat_wts[np.isinf(flat_wts)] = np.median(flat_wts[~np.isinf(flat_wts)])
    # test nan
    wts[np.isnan(wts)] = np.median(wts[~np.isnan(wts)])
    flat_wts[np.isnan(flat_wts)] = np.median(flat_wts[~np.isnan(flat_wts)])
    return wts, flat_wts

def kde_fit_quantiles(rtquants, nsamples=1000, bw=.1):
    """ takes quantile estimates and fits cumulative density function
    returns samples to pass to sns.kdeplot()
    """
    kdefit = KernelDensity(kernel='gaussian', bandwidth=bw).fit(rtquants)
    samples = kdefit.sample(n_samples=nsamples).flatten()
    return samples

def scurve_interpolate(x, y, kind='cubic'):
    interpol_fx = interp1d(x, y, kind=kind)
    xsim = np.linspace(x[0], x[-1], 10000, endpoint=True)
    ysim = interpol_fx(xsim)
    return xsim, ysim

def scurve_poly_fit(x, y, n=20):
    polysim = lambda p, pi, x, xi: p[pi]*x**xi
    ix = np.arange(n+1)
    poly_ix = zip(ix, ix[::-1])
    p = np.polyfit(x,y,n)
    xsim = np.linspace(x.min(), x.max(), 10000, endpoint=True)
    ysim = np.sum([polysim(p, pi, xsim, xi) for pi, xi in poly_ix],axis=0)
    return xsim, ysim

def fit_sigmoid(x, y):
    x = resize(x, lower=x.max(), upper=x.min())
    y = resize(y, lower=y.min(), upper=y.max())
    p0 = (np.mean(x), np.mean(y), .5, .5)
    p, pcov = optimize.leastsq(residuals, p0, args=(x, y), xtol=1.e-15, ftol=1.e-15, maxfev=20000)
    x0, y0, c, k = p
    xsim = np.linspace(x.min()-20, x.max()+20, 10000, endpoint=True)
    ysim = sigmoid(p, xsim)
    return xsim, ysim

def sigmoid(p, x):
    x0, y0, c, k = p
    y = c / (1 + np.exp(k * (x - x0))) + y0
    return y

def residuals(p, x, y):
    return y - sigmoid(p, x)

def resize(arr, lower=0.0, upper=1.0):
    arr = arr.copy()
    if lower > upper:
        lower, upper = upper, lower
    arr -= arr.min()
    arr *= (upper - lower) / arr.max()
    arr += lower
    return arr

def get_observed_vector(rt, quantiles=array([10, 30, 50, 70, 90])):
    """ takes array of rt values and returns binned counts (trials
    that fall between each set of quantiles in quantiles). also returns
    the total number of observations (len(rt)) and the RT values at those
    quantiles (rtquant)
    """
    inter_quantiles = array([quantiles[0] - 0] + [quantiles[i] - quantiles[i - 1] for i in range(1, len(quantiles))] + [100 - quantiles[-1]])
    rtquant = mq(rt, prob=quantiles * .01)
    ocounts = np.ceil((inter_quantiles) * .01 * len(rt)).astype(int)
    n_obs = np.sum(ocounts)
    return [ocounts, rtquant, n_obs]

def get_expected_vector(simrt, obsinfo):
    """ calculates the expected frequencies of responses for a
    set of simulated RTs, given. obsinfo is output of
    get_observed_vector() -->  [ocounts, rtquant, n_obs]
    simrt = pd.Series(simrt)
    """
    counts, q, n_obs = obsinfo[0], obsinfo[1], obsinfo[2]
    first = array([len(simrt[simrt.between(simrt.min(), q[0])]) / len(simrt)]) * n_obs
    middle = array([len(simrt[simrt.between(q[i - 1], q[i])]) / len(simrt) for i in range(1, len(q))]) * n_obs
    last = array([len(simrt[simrt.between(q[-1], simrt.max())]) / len(simrt)]) * n_obs
    expected = np.ceil(np.hstack([first, middle, last]))
    return expected

def ssrt_calc(df, avgrt=.3):
    dfstp = df.query('ttype=="stop"')
    dfgo = df.query('choice=="go"')
    pGoErr = ([idf.response.mean() for ix, idf in dfstp.groupby('idx')])
    nlist = [int(pGoErr[i] * len(idf)) for i, (ix, idf) in enumerate(df.groupby('idx'))]
    GoRTs = ([idf.rt.sort(inplace=False).values for ix,idf in dfgo.groupby('idx')])
    ssrt_list = ([GoRTs[i][nlist[i]] for i in np.arange(len(nlist))]) - avgrt
    return ssrt_list

def get_obs_quant_counts(df, quantiles=([.10, .30, .50, .70, .90])):
    if type(df) == pd.Series:
        rt = df.copy()
    else:
        rt = df.rt.copy()
    inter_quantiles = [quantiles[0] - 0] + [quantiles[i] - quantiles[i - 1] for i in range(1, len(quantiles))] + [1.00 - quantiles[-1]]
    obs_quant = mq(rt, prob=quantiles)
    observed = np.ceil((inter_quantiles) * len(rt) * .94).astype(int)
    return observed, obs_quant

def get_exp_counts(simdf, obs_quant, n_obs, quantiles=([.10, .30, .50, .70, .90])):
    if type(simdf) == pd.Series:
        simrt = simdf.copy()
    else:
        simrt = simdf.rt.copy()
    exp_quant = mq(simrt, prob=quantiles)
    oq = obs_quant
    expected = np.ceil(np.diff([0] + [pscore(simrt, oq_rt) * .01 for oq_rt in oq] + [1]) * n_obs)
    return expected, exp_quant


def pwts_group_error_calc(handler):
    """ get stdev across subjects (and any conds) in observedDF
    weight perr by inverse of counts for each resp. probability measure
    previously a bound method of dfhandler --> DataHandler class
    """
    groupedDF = handler.observedDF.groupby(handler.conds)
    perr = groupedDF.agg(np.nanstd).loc[:, handler.p_cols].values
    counts = groupedDF.count().loc[:, handler.p_cols].values
    nsplits = handler.nlevels * handler.nconds
    ndata = len(handler.p_cols)
    # replace stdev of 0 with next smallest value in vector
    perr[perr==0.] = perr[perr>0.].min()
    p_wt_bycount = perr * (1./counts)
    # set wts equal to ratio --> median_perr / all_perr_values
    pwts_ratio = np.nanmedian(p_wt_bycount, axis=1)[:, None] / p_wt_bycount
    # set extreme values to max_wt arg val
    pwts_ratio[pwts_ratio >= handler.max_wt] = handler.max_wt
    # shape pwts_ratio to conform to wtsDF
    idx_pwts = np.array([pwts_ratio]*handler.nidx)
    return idx_pwts.reshape(handler.nidx * nsplits, ndata)



def get_intersection(iter1, iter2):
    """ get the intersection of two iterables ("items in-common")
    """
    intersect_set = set(iter1).intersection(set(iter2))
    return ([i for i in intersect_set])



    #!/usr/local/bin/env python
from __future__ import division
from copy import deepcopy
import numpy as np
import pandas as pd
from numpy import array
from radd.models import Simulator
from radd.CORE import RADDCore
from radd.vis import plot_model_fits

class Model(RADDCore):
    """ Main class for instantiating, fitting, and simulating models.
    Inherits from RADDCore parent class (see CORE module).
    ::Arguments::
        data (pandas DF):
            data frame with columns 'idx', 'rt', 'acc', 'ttype', 'response',
            <Condition Name> declared in depends_on values
        kind (str):
            declares model type ['dpm', 'irace', 'pro']
            append 'x' to front of model name to include a dynamic
            bias signal in model
        inits (dict):
            dictionary of parameters (v, a, tr, ssv, z) used to initialize model
        fit_on (str):
            set if model fits 'average', 'subjects', 'bootstrap' data
        depends_on (dict):
            set parameter dependencies on task conditions
            (ex. depends_on={'v': 'Condition'})
        weighted (bool):
            if True (default), perform fits using a weighted least-squares approach
        quantiles (array):
            set the RT quantiles used to fit model
    """

    def __init__(self, data=pd.DataFrame, kind='xdpm', inits=None, fit_on='average', weighted=True, depends_on={'all':'flat'}, ssd_method=None, quantiles=np.arange(.1, 1.,.1)):

        super(Model, self).__init__(data=data, inits=inits, fit_on=fit_on, depends_on=depends_on, kind=kind, quantiles=quantiles, weighted=weighted, ssd_method=ssd_method)

    def optimize(self, plotfits=True, saveplot=False, keeplog=True, saveresults=True, saveobserved=False, custompath=None, sameaxis=False, progress=False):
        """ Method to be used for accessing fitting methods in Optimizer class
        see Optimizer method optimize()
        """
        if np.any([keeplog, saveplot, saveresults]):
            self.results_dir = self.handler.make_results_dir(custompath, get_path=True)
        if progress:
            self.make_progress_bars()
        for i in range(len(self.observed)):
            if self.track_subjects:
                self.pbars.update(name='idx', i=i, new_progress=self.idx[i])
            if not hasattr(self, 'init_params'):
                y, wts = self.iter_flat[i]
                self.set_fitparams(idx=i, y=y, wts=wts, nlevels=1, flat=True)
                finfo, popt, yhat = self.optimize_flat(progress=progress)
                self.init_params = deepcopy(popt)
            else:
                popt = self.init_params
            if not self.is_flat:
                y, wts = self.iter_cond[i]
                self.set_fitparams(idx=i, y=y, wts=wts, nlevels=self.nlevels, flat=False)
                finfo, popt, yhat = self.optimize_conditional(p=popt)
            self.assess_fit(finfo, popt, yhat, keeplog)
            if plotfits:
                self.plot_model_fits(y=y, yhat=yhat, save=saveplot, sameaxis=sameaxis)
        if progress and not self.is_nested:
            self.pbars.clear()
        if saveresults:
            self.write_results(saveobserved)

    def optimize_flat(self, progress=False):
        """ optimizes flat model to data collapsing across all conditions
        ::Arguments::
            p (dict):
                parameter dictionary to initalize model, if None uses init params
                passed by Model object
        ::Returns::
            yhat_flat (array): model-predicted data array
            finfo_flat (pd.Series): fit info (AIC, BIC, chi2, redchi, etc)
            popt_flat (dict): optimized parameters dictionary
        """
        pbars=None
        if not self.finished_sampling:
            self.sample_param_sets()
        inits, globalmin = self.filter_param_sets()
        if progress:
            self.pbars.reset_bar('glb_basin', init_state=globalmin)
            pbars = self.pbars
        # Global Optimization w/ Basinhopping (+TNC)
        p = self.opt.hop_around(inits=inits, pbars=pbars)
        # Flat Simplex Optimization of Parameters at Global Minimum
        finfo, popt, yhat = self.opt.gradient_descent(p=p)
        return finfo, popt, yhat

    def optimize_conditional(self, p=None):
        """ optimizes full model to all conditions in data
        ::Arguments::
            p (dict): parameter dictionary to initalize model, if None uses init params
        ::Returns::
            yhat (array): model-predicted data array
            finfo (pd.Series): fit info (AIC, BIC, chi2, redchi, etc)
            popt (dict): optimized parameters dictionary
        """
        if p is None:
            p = self.__check_inits__(self.init_params)
        # Pretune Conditional Parameters
        p, fmin = self.opt.run_basinhopping(p)
        # Final Simplex Optimization
        finfo, popt, yhat = self.opt.gradient_descent(p=p)
        return finfo, popt, yhat

    def assess_fit(self, finfo, popt, yhat, keeplog=False):
        """ wrapper for analyze.assess_fit calculates and stores
        rchi, AIC, BIC and other fit statistics
        """
        fp = deepcopy(self.fitparams)
        fp['yhat'] = yhat.flatten()
        y = fp['y'].flatten()
        wts = fp['wts'].flatten()
        # fill finfo dict with goodness-of-fit info
        finfo['chi'] = np.sum(wts * (fp['yhat'] - y)**2)
        finfo['ndata'] = len(fp['yhat'])
        finfo['df'] = finfo.ndata - finfo.nvary
        finfo['rchi'] = finfo.chi / finfo.df
        finfo['logp'] = finfo.ndata * np.log(finfo.rchi)
        finfo['AIC'] = finfo.logp + 2 * finfo.nvary
        finfo['BIC'] = finfo.logp + np.log(finfo.ndata * finfo.nvary)
        self.finfo = finfo
        self.popt = popt
        self.yhat = yhat
        if keeplog:
            self.log_fit_info(finfo, popt, fp)
        self.fill_yhatDF(yhat=yhat, fitparams=fp)
        self.fill_fitDF(finfo=finfo, fitparams=fp)

    def optimize_nested_models(self, models=[], saveplot=True, plotfits=True, progress=False, keeplog=True):
        """ optimize externally defined models in model_list using same init parameters
        as the current model for conditional fits (NOTE: only works with fit_on='average')
        """
        self.is_nested = True; self.nmodels=len(models)
        if not hasattr(self, 'init_params'):
            self.optimize(plotfits=plotfits, progress=progress)
        for i, pdep in enumerate(models):
            if progress:
                self.pbars.update(name='models', i=i)
            self.set_fitparams(depends_on = {pdep: list(self.clmap)[0]})
            p = self.__check_inits__(deepcopy(self.init_params))
            finfo, popt, yhat = self.optimize_conditional(p=p)
            self.assess_fit(finfo, popt, yhat, keeplog)
            if plotfits:
                self.plot_model_fits(self.fitparams.y, yhat, self.fitparams, save=saveplot)
        if progress:
            self.pbars.clear()

    def simulate(self, p=None, analyze=True):
        """ simulate yhat vector using
        :: Arguments ::
            p (dict):
                parameters dictionary
            analyze (bool):
                if True (default) returns yhat vector. else, returns decision traces
        :: Returns ::
            out (array):
                1d array if analyze is True, else ndarray of decision traces
        """
        if p is None:
            p = self.inits
        return self.opt.simulator.sim_fx(p, analyze=analyze)
